;;; RTGParser: is a CKY parser for regional tile grammars; It uses the algorithm
;;; described in: 
;;;
;;; A. Cherubini, S. Crespi Reghizzi, M. Pradella, Regional languages and
;;; tiling: a unifying approach to picture grammars, MFCS 2008, volume 5162
;;; of Lecture Notes in Computer Science, pages 253-264, 2008
;;; (http://home.dei.polimi.it/pradella/papers/region_cr_rev.pdf)
;;;
;;; Version 081214
;;;
;;; Written in PLT Scheme (v. 4): http://plt-scheme.org/
;;; by Luca Solbiati (lucajdvr@gmail.com)
;;; (with modifications by Matteo Pradella)
;;;
;;; --------------------------------------------------------------------------
;;;
;;; Copyright (C) 2008
;;; Luca Solbiati   (lucajdvr@gmail.com)
;;; Matteo Pradella (pradella@elet.polimi.it)
;;;
;;; This program is free software; you can redistribute it and/or modify
;;; it under the terms of the GNU General Public License as published by
;;; the Free Software Foundation; either version 2 of the License, or
;;; (at your option) any later version.
;;;
;;; This program is distributed in the hope that it will be useful,
;;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;;; GNU General Public License for more details.
;;;
;;; You should have received a copy of the GNU General Public License
;;; along with this program; if not, write to the Free Software
;;; Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
;;;
;;; --------------------------------------------------------------------------
;;; ----------------------------------------------------------------------------

#lang scheme/base

(provide
 parse!
 set-rules!
 toggle-trace!)

(require 
  (file "mazpa-utils.scm")
  srfi/25 scheme/list srfi/69)

;; -----------------------------Global definitions------------------------------

(define *fs-rules*           '()) ;Fixed size rules
(define *vs-rules*           '()) ;Variable size rules
(define *nt-list*            '()) ;The nonterminals list
(define *recognition-matrix* '()) ;The recognition matrix
(define *picture*            '()) ;The picture to be parsed
(define *trace*               #f)


;; -----------------------------------------------------------------------------
;;                                 Algorithm
;; -----------------------------------------------------------------------------

;; This function enable (disable) the parsing trace
(define (toggle-trace!)
  (set! *trace* (not *trace*)))

;; This function sets the grammar rules for the parser: *fs-rules* is a list of
;; list containing the terminal character as first element and than all the
;; nonterminals; *vs-rules* is a list of vector containing as first element
;; the nonterminal, as second terminal the list of pairs of nonterminals in
;; horizontal adjacency relationship and as third element the list of pairs
;; of nonterminals in vertical adjacency relationship.
;; The function also fills up the nonterminals list
(define (set-rules! fs-rules vs-rules)
  (set! *fs-rules* fs-rules)
  (for-each
   (lambda (nt)
     (for-each
      (lambda (w)
        (set! *vs-rules* (cons (vector (car nt) (hw w) (vw w)) 
                               *vs-rules*)))
      (cdr nt))
     (set! *nt-list* (cons (car nt) *nt-list*))) 
   vs-rules)
  (for-each
   (lambda (r)
     (for-each
      (lambda (nt)
        (set! *nt-list* (cons nt *nt-list*)))
      (cdr r)))
   fs-rules)
  (set! *nt-list* (remove-duplicates *nt-list*)))

;; This function given a regional set of tiles w returns the list of pairs
;; (N1 N2) such that N1 is in horizontal adjacency relationship with N2
(define (hw w)
  (define (hw-aux w i j)
    (define (hw-elem w i j)
      (let ((N1 (list-ref (list-ref w i) j))
            (N2 (list-ref (list-ref w i) (add1 j))))
        (cond [(not (eq? N1 N2))
               (vector N1 N2)])))
    (cond [(< j (- (length (car w)) 2))
           (cons-no-void (hw-elem w i j) (hw-aux w i (add1 j)))]
          [(and (= j (- (length (car w)) 2)) (< i (sub1 (length w))))
           (cons-no-void (hw-elem w i j) (hw-aux w (add1 i) 0))]
          ((and (= j (- (length (car w)) 2)) (= i (sub1 (length w))))
           (if (void? (hw-elem w i j)) '()
               (list (hw-elem w i j))))))
  (remove-duplicates (hw-aux w 0 0)))

;; This function given a regional set of tiles w returns the list of pairs
;; (N1 N2) such that N1 is in vertical adjacency relationship with N2
(define (vw w)
  (define (vw-aux w i j)
    (define (vw-elem w i j)
      (let ((N1 (list-ref (list-ref w i) j))
            (N2 (list-ref (list-ref w (add1 i)) j)))
        (cond [(not (eq? N1 N2))
               (vector N1 N2)])))
    (cond [(< i (- (length w) 2))
           (cons-no-void (vw-elem w i j) (vw-aux w (add1 i) j))]
          [(and (= i (- (length w) 2)) (< j (sub1 (length (car w)))))
           (cons-no-void (vw-elem w i j) (vw-aux w 0 (add1 j)))]
          ((and (= i (- (length w) 2)) (= j (sub1 (length (car w)))))
           (if (void? (vw-elem w i j)) '()
               (list (vw-elem w i j))))))
  (remove-duplicates (vw-aux w 0 0)))

;; This function initializes the recognition matrix
(define (initialize-rm!)
  (let ((m (length *picture*))
        (n (length (car *picture*))))
    (set! *recognition-matrix* (make-array (shape 0 m 0 n 0 m 0 n)))))

;; This function set the nonterminal list nt-list at position (i,j,k,l) in the
;; recognition matrix
(define (rm-set! nt-list i j k l)
  (array-set! *recognition-matrix* i j k l 
              (remove-duplicates (append nt-list (rm-ref i j k l)))))

;; This function get the nonterminal list at position (i,j,k,l) in the
;; recognition matrix
(define (rm-ref i j k l)
  (let ([rm-elem (array-ref *recognition-matrix* i j k l)])
    (if (list? rm-elem) rm-elem '())))

;; This function returns the subdomains vector as an hash table, whose keys are
;; the nonterminals and values are lists of subdomain represented as a vector
(define (computeD i j k l)
  (let ((D (make-hash))
        (res '(())))
    (for-range-2 i k j l
      (lambda (i1 j1)
        (for-range-2 i1 k j1 l
          (lambda (k1 l1)
            (for-each (lambda (N)
                        (hash-set! D N
                                   (let ((d\N (hash-ref D N #f)))
                                     (if d\N
                                         (cons (cons N (vector i1 j1 k1 l1)) d\N)
                                         (list (cons N (vector i1 j1 k1 l1)))))))
            (rm-ref i1 j1 k1 l1))))))
    (for-each
     (lambda (N)
       (when (not (hash-ref D N #f))
         (hash-set! D N (list (cons N #(-1 -1 -1 -1))))))
     *nt-list*)
    (cartprod (hash-map D (lambda (N set) set)))))

;; Given Hw and Vw and a subdomain (i,j,k,l) if exists a rule C->w such that
;; (p0,pi0)=>G(p1,pi1), where d is in pi0 and spict(p0,d) is a C-picture
(define (check-rule? Hw Vw i j k l)
  (ormap
    (lambda (D)
      (and
	(andmap
	  (lambda (N)
	    (let ((Na (vector-ref N 0))
		  (Nb (vector-ref N 1)))
	      (cond
		[(and (eq? Na '$) (not (eq? Nb '$)))
		 (= j (vector-ref (cdr (assq Nb D)) 1))]
		[(and (not (eq? Na '$)) (eq? Nb '$))
		 (= l (vector-ref (cdr (assq Na D)) 3))]
		[else
		  (let ((da (cdr (assq Na D)))
			(db (cdr (assq Nb D))))
		    (and
		      (= (vector-ref db 1) (add1 (vector-ref da 3)))
		      (>= (vector-ref db 2) (vector-ref da 0))
		      (>= (vector-ref da 2) (vector-ref db 0))))]))) Hw)
	(andmap
	  (lambda (N)
	    (let ((Na (vector-ref N 0))
		  (Nb (vector-ref N 1)))
	      (cond
		[(and (eq? Na '$) (not (eq? Nb '$)))
		 (= i (vector-ref (cdr (assq Nb D)) 0))]
		[(and (not (eq? Na '$)) (eq? Nb '$))
		 (= k (vector-ref (cdr (assq Na D)) 2))]
		[else
		  (let ((da (cdr (assq Na D)))
			(db (cdr (assq Nb D))))
		    (and
		      (= (vector-ref db 0) (add1 (vector-ref da 2)))
		      (>= (vector-ref db 3) (vector-ref da 1))
		      (>= (vector-ref da 3) (vector-ref db 1))))]))) Vw)))
    (computeD i j k l)))
             

;; This function applies all the fixed size rules
(define (apply-fs-rules)
  (for-range-2 0 (sub1 (length *picture*)) 0 (sub1 (length (car *picture*)))
               (lambda (i j)
                 (rm-set!
                  (cdr (assoc (list-ref (list-ref *picture* i) j)
                              *fs-rules*)) i j i j))))

;; This function applies all the variable size rules
(define (apply-vs-rules)
  (let ([m (length *picture*)]
        [n (length (car *picture*))])
    (for-range 1 (* m n)
      (lambda (size)
        (for-range-2 1 m 1 n
          (lambda (h w)
            (when (= (* h w) size)
              (for-range-2 0 (sub1 m) 0 (sub1 n)
                (lambda (i j)
                  (let ([k (sub1 (+ i h))]
                        [l (sub1 (+ j w))])
                    (when (and (< k m) (< l n))
                      (begin
                        (when *trace*
                          (begin
                            (display (list i j k l))
                            (newline)
                            ))
                        (for-each (lambda (n)
                                    (let ((N (vector-ref n 0))
                                          (Hw (vector-ref n 1))
                                          (Vw (vector-ref n 2)))
                                      (when (check-rule? Hw Vw i j k l)
                                        (begin
                                          (rm-set! (list N) i j k l)
                                          (when *trace*
                                            (begin
                                              (display N)
                                              (newline)))))))
                                  *vs-rules*)))))))))))))


;; This function start the parsing of the given picture
(define (parse! picture)
  (set! *picture* picture)
  (initialize-rm!)
  (apply-fs-rules)
  (apply-vs-rules)
  (display (if (member 'S (rm-ref 0 0 
                                  (sub1 (length picture))
                                  (sub1 (length (car picture)))))
               "Yes" "No"))
  (newline))

